"""Toolbox for Data structures on Canadian climate data CDs.
Copyright (C) 1999 Bernhard Reiter 
<bernhard@ffii.org> http://www.uwm.edu/~bernhard/

Experimental Collection of functions, struct format strings and classes to
access climate data on the "Environment Canada" CD, called
"Canadian Daily Climate Data Temperature and Precipitation"
tested with "Western Canada 1997" and "Eastern Canada 1997", copyright 1999.

Format as in format description "CD2 File Format; March, 1993" and
some assumptions.

	License:
	--------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
This source is Free Software under the GNU GPL License (see gpl.txt).


"""
__version__=0.6
# 24.8.1999 Bernhard Reiter
#	initial
# 25.8.1999 Bernhard
#	added docstrings
#	changed a lot of functions
#	improved weather_station class
# 26.8.1999 Bernhard
#	improved and changed weather_station class
# 27.8.1999 Bernhard
#	corrected Mark's outformat a space was need in all, but the first line
#	for each year
# 28.8.1999 Bernhard
#	improved documentation
#	summaries are unpacked and returned by weather_class.get_data() now
#	weather_class.write_temperature_for_mark() works different internally 
#	added: weather_class_get_data()

import sys
import os.path	# portable path manipulations

import array	# Efficient arrays of uniformly typed numeric values.
import struct	# Interpret strings as packed binary data

# from the Library Reference documentation for struct:
#  Format 
#	C Type 		Python
#
#    x pad byte 	no value 
#    c char 		string of length 1 
#    b signed char 	integer 
#    B unsigned char 	integer 
#    h short 		integer 
#    H unsigned short 	integer 
#    i int 		integer 
#    I unsigned int 	integer 
#    l long 		integer 
#    L unsigned long 	integer 
#    f float 		float 
#    d double 		float 
#    s char[] 		string 
#    p char[] 		string 
#    P void * 		integer 

################################################################################
# Format strings for the struct pack and unpack functions
# corresponding to the binary format within the data and index files
#
Index_File_header_fmt= ( "<" + # indicates little-endian data
	"4s" + 		# ID
	"HH" +		# MaxLat, MinLat
	"HH" +		# MaxLong, MinLong
	"hh" +		# MaxElev, MinElev
	"HH" +		# EarliestYear, LatestYear
	"H"  +		# Number of Stations in the district
	"3s" +		# District ID
	"42s"		# District Name
	)

Index_File_data_fmt=  ( "<" + # indicates little endian data
	"4s" +		# CSN, the last four characters of the 7-char statHon ID
	"24s" +		# StationName
	"3s" +		# Airport
	"HH" +		# Lat, Long
	"h"  +		# Elevation
	"7H" +		# First Year
	"7H" +		# Last Year
	"H"		# [20] Starting Record Number
	)


Data_File_header_fmt= ( "<" +
	"4s" +		# Data file version id (should be "WWWW")
	"7s" +		# CSN = station ID number
	"24s" +		# Station Name
	"3s" +		# Airport
	"HH" +		# Lat, Long
	"H"  +		# Elevation
	"HH" +		# First Year, Last Year    for station
	"300H" +	# Record Num for each year's 1st element starting w 1801
	"300B" +	# Available indicator, left to right 6 bit of each byte
	"123x" )	# 

Data_File_data_fmt= (  "<" + 
	"732s" +	# contains array of short ints: one day's value
	"183s" +	# contains array of nybbles: one day's flag
	"156s" )	# contains 12 month summaries a 13 bytes

Record_size=struct.calcsize(Data_File_data_fmt)
if Record_size != struct.calcsize(Data_File_header_fmt) or Record_size != 1071:
    sys.stderr("Something wrong with the internal record structure formats!\n")
    sys.exit(5)
    

Maximum_Temperature_fmt = (  "<" + 
        "12h"  +	#  Mean Max for Month 
        "12B"  +	#  # Days Missing 
        "12B"  +	#  # Days with Max above 0  
        "12h"  +	#  Maximum Max for Month 
        "12B"  +	#  Date of Maximum Max 
        "12h"  +	#  Minimum Max for Month 
        "12B"  +	#  Date of Minimum Max 
        "12B"  +	#  Max # Days in a Row with Missing Data 
        "12B"  +	#  # Days with Misg Mean Temperature 
        "12B"  )	#  Maximum # Days in a row with Misg Mean Temperature 

Minimum_Temperature_fmt = ( "<" +
        "12h"  +	# Mean Minimum for Month 
        "12B"  +	# # Days Missing 
        "12B"  +	# # Days with Min above 0  
        "12h"  +	# Maximum Min for Month 
        "12B"  +	# Date of Maximum Min 
        "12h"  +	# Minimum Min for Month 
        "12B"  +	# Date of Minimum Min 
        "12B"  +	# Max # Days in a Row with Missing Data 
        "12h"  )	# Mean Temperature 

Precipitation_fmt = ( "<" + 
        "12H"  +	# Total for Month 
        "12B"  +	# # Days Missing 
        "12B"  +	# # Days with > 0 cm/mm 
        "12H"  +	# Max one-day total for Month 
        "12B"  +	# Date of Max 
        "12B"  +	# # Days with > Trace 
        "12B"  +	# # Days with > 1 cm 
        "12B"  +	# # Days with Uncrtn/accum pcpn 
        "12B"  +	# Maximum # Days in a row with Uncrtn/accum Pcpn 
        "24x" )

Snow_on_the_Ground_fmt = ( "<" + 
        "12H"  +	# Median for Month 
        "12B"  +	# # Days Missing 
        "12B"  +	# Days with >0 cm 
        "12H"  +	# Max for Month 
        "12B"  +	# Date of Max 
        "12H"  +	# Min for Month 
        "12B"  +	# Date of Min 
        "12B"  +	# Days with > Trace 
        "12B"  +	# Days with > 1 cm 
        "12x" )

################################################################################
#
#

def read_all_records(file,fmt):
    """Return list containing all records read from the monotoniously build file.
    
    	Parameter:
    		file	an open file object (seeked to the right place)
    		fmt 	a struct module format string
    	
    	This routine will read till end of the file and 
    	give a message if the last read count was not 0.
    """
    list=[]
    size=struct.calcsize(fmt)

    while 1:
        s=file.read(size)
	if len(s) == size:
		list.append(struct.unpack(fmt,s))
		continue

	if len(s)!=0:
		sys.stderr.write("Only got %d bytes on last read!" % len(s))
	return list



def build_file_index(topdirectory):
    """ Returns list of (index filename, data filename, district) tuples.
    
    Reads the "index.all" file in the topdirectory and calculates all other
    index filenames.
    """
    file=open(os.path.join(topdirectory,"index.all"),"rb")
    list=read_all_records(file,Index_File_header_fmt)
    file.close()

    # Okay, using a lambda construct and map is a bit complicated, but elegant
    def create_index_name(topdir):
    	return lambda x,td=topdir: \
	    ( os.path.join(td,x[10][0],"Index."+x[10]),
	    os.path.join(td,x[10][0],"Data."+x[10]),
	    x[10])

    return map(create_index_name(topdirectory),list)


################################################################################
#
class weather_station:
    """Manages data for one station.

    
    To creat it, you need to give it the 
    filetuple with (Indexfilename,datafilename,districtnumber) and
    the index record. A function to do this is get_station_list().
    The datafile is not open, when the object is instanciated.

    Call	get_data(year,what)  what being among avail.keys()
    			to get a tuple with two lists: values, flags, both raw

		write_temperature_for_mark(.... )

		deactivate() to close the open datafile, if open.

		AvailFor(year)	get the start_record and the avail_flags
				for the year

		decode_avail_flags(avail_flags) to get text representation

    Variables (see __init__()), most visible:
    	stationnumber
	district
	dataindex	if activate() was called contains the tuple
			with all the data of the data file header.

    Internal:

    Class constant:
    	avail	dictionary	maps string representation to
				a tuple with
				(avail_flag, denominator, summary binary format)
    """

    avail= { 
    "MAX_TEMP":(1,10.0,Maximum_Temperature_fmt),
    "MIN_TEMP":(1<<1,10.0,Minimum_Temperature_fmt),
    "ONE_DAY_RAIN":(1<<2,10.0,Precipitation_fmt),
    "ONE_DAY_SNOW":(1<<3,10.0,Precipitation_fmt),
    "ONE_DAY_PRECIPITATION":(1<<4,10.0,Precipitation_fmt),
    "SNOW_ON_THE_GROUND":(1<<5,10.0,Snow_on_the_Ground_fmt)
    }
# flags explanation out of cdcd.doc, values out of format description
#  0   (no flag)
#  7-13 unused
# 14   29th of February in non leap years
#  5 A Amount accumulated over more than one day; Previous value's flag was
#    C or L
#  3 C Precipitation occurred; amount is uncertain; recorded value is 0;
#      value displayed is the word "Yes"
#  1 E Estimated
#  6 F Amount accumulated over more than one day and estimated
#  4 L Precipitation may or may not have occurred; amount is unknown;
#      recorded value is 0; value displayed is the word "Maybe"
# 15 M Missing
#  2 T A trace occurred; recorded value is zero


    
    def __init__(self,filetuple,index_record):
        """Initialise instance variables."""
	self.index_record=index_record
	self.indexfilename=filetuple[0]
	self.datafilename=filetuple[1]
	self.district=filetuple[2]
	self.startrecord=self.index_record[20]
	self.stationnumber=self.district+self.index_record[0]
	self.datafileopen=0
	self.dataindex=[]

    def activate(self):
        """Make sure the datafile is open and we have the dataindex."""
    	if not self.datafileopen:
		self.datafile=open(self.datafilename,"rb")
		self.datafileopen=1
	if not self.dataindex:
		self.datafile.seek((self.startrecord-1)*Record_size)
		string=self.datafile.read(Record_size)
		self.dataindex=struct.unpack(Data_File_header_fmt,string)


    def deactivate(self):
        """Close the datafile if necessary."""
    	if self.datafileopen:
		self.datafile.close()
		self.datafileopen=0

    def AvailFor(self,year):
        """Return tuple with recordindex and availabilty flags or null."""
	self.activate()

    	recnumb_index=year-1801+9
	dataavail_index=year-1801+309

	firstrec= self.dataindex[recnumb_index]
	if firstrec:
		return(firstrec, self.dataindex[dataavail_index])
	else:
		return 0


    def get_data(self,year,what):
        """Returns list w [values,flags,summaries] for wanted year and item.

	All non-missing values have been turned into normal representation.
	The missing values are undefined.
	"""
	
    	tuple=self.AvailFor(year)
	if (not tuple) or (not self.avail[what][0] & tuple[1]):
		return 0

	# seek to the right datablock
	i=0
	while not self.avail[what][0] & (1<<i) and i<=5:
	    if (1<<i) & tuple[1]:
		i=i+1
	self.datafile.seek((self.startrecord+tuple[0]-1+i)*Record_size)

	# get datablock, unpack it
	string=self.datafile.read(Record_size)
	r=struct.unpack(Data_File_data_fmt,string)

	raw_values=array.array("h")
	raw_values.fromstring(r[0])
	flags=array.array("B")
	for i in range(183):
		flags.append(ord(r[1][i])>>4&15)
		flags.append(ord(r[1][i])&15)

	# sanity check:
	if len(raw_values)!=len(flags):
		sys.stderr.write("Arg! Something is very wrong! :-( \n")
		sys.exit(10)

	# transform values into normal representation
	def trans_into_value(denominator):
		return lambda v, d=denominator: v/d

	values=array.array("f")
	values.fromlist(map(trans_into_value(self.avail[what][1]),raw_values))

	# unpack summaries
	summaries=struct.unpack(self.avail[what][2],r[2])

	return [values,flags,summaries]


    def write_temperature_for_mark(self, file, year):
        """Appends daily max and min temperatures to file in Mark S's format.
	
	This is a special format used by another older program.
	It uses Fahrenheit (yuck) and non-leap years have a missing value
	at the end.
	"""

	def fix_leap_year(raw):
	    """ pay attention to leap year at position 31+29=60. """
	    if raw[1][59]==14:
	        #shift leap year to end ->written out like missing value later
		raw[0].append(raw[0][59])
		raw[0]=raw[0][:59]+raw[0][60:]
		raw[1].append(raw[1][59])
		raw[1]=raw[1][:59]+raw[1][60:]

	def trans_into_fahrenheit_or_missing(celsius,flag):
		if flag in [0,1]:
			return (9.0/5.0)*celsius+32
		else:
			return 999.0

	def write_in_marks_format(list,file=file,self=self,year=year):
        	file.write("%7s%4d" % (self.stationnumber, year) )
        	for i in range(0,366,20):
		    # writing one line
		    if not i==0:
		        file.write(" ")
        	    for t in list[i:i+20]:
			#we have to be sure to write only three characters
			if t<-99:
			    t=-99
			    sys.stderr.write(
			    	"Temperature below -99 detected and cut!\n")
			if t>999:
			    t=999
			    sys.stderr.write(
			    	"Temperature value >999 detected and cut!\n")
        	        file.write("%3.0f" % t)
        	    file.write("\n")


	rawmax=self.get_data(year,"MAX_TEMP")
	rawmin=self.get_data(year,"MIN_TEMP")
	if not (rawmax and rawmin):
		sys.stderr.write("%s's Temp not available for s %s\n" % 
			(year, self.stationnumber) )
		return 0
	fix_leap_year(rawmax)
	fix_leap_year(rawmin)

	max=map(trans_into_fahrenheit_or_missing,rawmax[0],rawmax[1])
	min=map(trans_into_fahrenheit_or_missing,rawmin[0],rawmin[1])

	del(rawmax)
	del(rawmin)

	write_in_marks_format(max)
	write_in_marks_format(min)
#
# END of class 
################################################################################


def decode_avail_flags(avail_flags):
    """ Returns unordered list of text representations of avail flags. 
    
    Needs the avail_flags integer value as argument.
    """
    items=[]
    for repr in weather_station.avail.keys():
        if weather_station.avail[repr][0] & avail_flags:
            items.append(repr)
    return items


def get_station_list(topdirectory):
    """ Return list of weather_station objects."""

    files=build_file_index(topdirectory)
    wlist=[]
    
    for filetuple in files:
	# read index file
        file=open(filetuple[0],"rb")
	string=file.read(struct.calcsize(Index_File_header_fmt))
	Index_File_header=struct.unpack(Index_File_header_fmt,string)

	Index_File_data_list=read_all_records(file,Index_File_data_fmt)
	if len(Index_File_data_list) != Index_File_header[9]:
		sys.stderr.write("Found %d out of %d!\n" % 
			(len(Index_File_data_list), Index_File_header[9]) )
        file.close()

	for record in Index_File_data_list:
            wlist.append(weather_station(filetuple,record))

    return wlist

